/*
 *	PAL software decoder
 *	includes Y/C separation and YCrCb to RGB conversion
 *
 *	Copyright (C) 1999, Ewald Snel
 *
 *WWW		http://esnel.op.het.net/
 *e-mail	esnel@cistron.nl
 */

#include <stdio.h>
#include <math.h>
#include <time.h>
#include "esneldecode.h"


#define decWidth	370	// decoded (display/output) width
#define decHeight	240	// decoded (display/output) height
#define rawWidth	2268		// input (raw) width
#define rawOffset	5*rawWidth - 102	// offset of first pixel

// decode (static)
#define Pi			3.14159265358979323846
#define divPi		(0x80000000L /Pi)
#define FcHF0 		((int) (227.25L * 0x100000000L /rawWidth))
#define FcVF0 		((int) (  -.25L * 0x100000000L))


static unsigned char	scanLuma[384];
static char				scanCbU[192], scanCrV[192];
static int				cosint[1024];
static bool				isInitialized = false;

static unsigned char clip8[1024];
static int PvU[decWidth/2], PvV[decWidth/2];


static void InitializeDecoder();
static void GetColorBurstData( unsigned char *src, unsigned int *cbPhase,
  unsigned int *cbHF, unsigned int *cbVF, unsigned int *cbOdd );
static void GetLumaSamples( unsigned char src[], bool lladjust, int brightness, int contrast );
static void GetChromaSamples( unsigned char src[],
							  unsigned int cbPhase, unsigned int cbHF, unsigned int cbVF,
							  int saturation, int vsaturation );
static void SetPixels(int y);

static long show_settings_timeout;


static bool adjust_luma_level = false;

static struct
{
	int brightness;
	int contrast;
	int saturation;
} color_cfg[2] = { {87, 174, 245}, {170, -181, 245} };

static int color_cfg_i;


EsnelDecoder::EsnelDecoder()
{
 	_description = "Improvisado PAL Color";
   _outwidth = decWidth;
   _outheight = decHeight;
	_use_vbi = true;
}

void EsnelDecoder::processKey(WPARAM key)
{
	bool show_it = false;

	switch(key)
   {
   	case VK_PRIOR:
        	show_it = true;
      	if (color_cfg[color_cfg_i].brightness < 511)
				color_cfg[color_cfg_i].brightness+=2;
         break;
      case VK_NEXT:
        	show_it = true;
      	if (color_cfg[color_cfg_i].brightness > -511)
				color_cfg[color_cfg_i].brightness-=2;
         break;
   	case VK_HOME:
        	show_it = true;
      	if (color_cfg[color_cfg_i].contrast < 511)
				color_cfg[color_cfg_i].contrast+=2;
         break;
      case VK_END:
        	show_it = true;
      	if (color_cfg[color_cfg_i].contrast > -511)
				color_cfg[color_cfg_i].contrast-=2;
         break;
   	case VK_INSERT:
        	show_it = true;
      	if (color_cfg[color_cfg_i].saturation < 511)
				color_cfg[color_cfg_i].saturation+=2;
         break;
      case VK_DELETE:
        	show_it = true;
      	if (color_cfg[color_cfg_i].saturation > -511)
				color_cfg[color_cfg_i].saturation-=2;
         break;

		case 'A':
      	adjust_luma_level = !adjust_luma_level;
         break;
	
		case VK_F3:
			color_cfg[color_cfg_i].saturation=-color_cfg[color_cfg_i].saturation;
		break;
   }

   if (show_it)
   	show_settings_timeout = time(0) + 5;
}

const char* EsnelDecoder::getHelp(void) const
{
	return("  PG UP/DOWN: Ajustar brillo\n"
          "  HOME/END: Ajustar contraste\n"
          "  INSERT/DELETE: Ajustar saturacion\n"
          "  A: Ajustar nivel de brillo segun la seal de color\n"
          );
}


/*
 Decode next field of PAL data
*/
void EsnelDecoder::decodeFrame(unsigned char *src, int offset,
  bool invert_polarity)
{
	bool lladjust = adjust_luma_level;
	color_cfg_i = (invert_polarity)?1:0;
	int brightness = color_cfg[color_cfg_i].brightness;
	int contrast = color_cfg[color_cfg_i].contrast;
	int saturation = color_cfg[color_cfg_i].saturation;

	unsigned int cbPhase, cbHF, cbVF, cbOdd;
	int i, y;

	if (!isInitialized)
		InitializeDecoder();

	src += offset;
	GetColorBurstData( src, &cbPhase, &cbHF, &cbVF, &cbOdd );

	src += getColorBurstWidth();

	brightness += (contrast < 0) ? 85 : 0;
	cbHF	   *= 10;

	lockDisplay();
      
	for (y=0; y < decHeight; y+=2)
	{
		GetChromaSamples( src, cbPhase, cbHF, cbVF, saturation, (cbOdd ? -saturation : saturation) );

		for (i=0; i < 2; i++)
		{
			GetLumaSamples( src, lladjust, brightness, contrast );
			SetPixels(y+i);

			src += rawWidth;
		}
		cbPhase	+= 2*cbVF;
	}

   unlockDisplay();

   if (show_settings_timeout > time(0))
   {
   	char buf[500];
	   sprintf(buf, "Brillo %d\nContraste %d\nSaturacion %d",
        color_cfg[color_cfg_i].brightness,
        color_cfg[color_cfg_i].contrast,
        color_cfg[color_cfg_i].saturation);
	   drawText(70, 120, buf, RGB(255, 255, 255));
   }
}


/*
 Initialize decoder
*/
static void InitializeDecoder()
{
	int i;

	for (i=0; i < 1024; i++)
	{
		cosint[i] = (int) (0x7fffffff * cos( i*Pi /512.0 )) >> 16;
		clip8[i]  = (i < 320) ? 0 : ((i > 320+255) ? 255 : i-320);
	}
	isInitialized = true;
}


static void SetPixels(int y)
{
	int i, y0, y1, u, v, cr, cg, cb;
	RGBTRIPLE pixels[decWidth];

	for (i=0; i < decWidth/2; i++)
	{
			u  = scanCbU[i];
			v  = scanCrV[i];
			y0 = scanLuma[i+i];
			y1 = scanLuma[i+i+1];
			cr = (64*320+32 +         73*v) >> 6;
			cg = (64*320+32 -  25*u - 37*v) >> 6;
			cb = (64*320+32 + 130*u       ) >> 6;

			pixels[i*2].rgbtRed = clip8[y0 + cb];//clip8[y0 + cr];
         pixels[i*2].rgbtGreen = clip8[y0 + cg];
         pixels[i*2].rgbtBlue = clip8[y0 + cr];//clip8[y0 + cb];

			pixels[i*2+1].rgbtRed = clip8[y1 + cb];//clip8[y1 + cr];
         pixels[i*2+1].rgbtGreen = clip8[y1 + cg];
         pixels[i*2+1].rgbtBlue = clip8[y1 + cr];//clip8[y1 + cb];
	}

	drawPixels(0, y, pixels, decWidth);
}


/*
 Decode single scanline of Raw 8bit samples in PAL format to grayscale
 luma (Y) samples at half PAL (50%) resolution.
*/
static void GetLumaSamples( unsigned char src[], bool lladjust, int brightness, int contrast )
{
	int x, l;

	// adjust luma level to color burst
	if (lladjust)
	{
		int cbluma = 0;

		for (x=0; x < 32; x++)
		{
			cbluma += src[x - 192];
		}
		brightness += ((384 - cbluma) * contrast) >> 11;
	}

	// luma notchfilter, remove color frequency, downsample
	for (x=0; x < decWidth; x++)
	{
		l  = brightness;
		l += ((src[-3]+src[-2]+src[-1]+src[ 0] + src[1]+src[2]+src[3]+src[4]) * contrast) >> 9;
		scanLuma[x] = (l < 0) ? 0 : ((l > 255) ? 255 : l);
		src += 5;
	}
}


/*
 Decode single scanline of Raw 8bit samples in PAL format to chroma (U,V)
 samples at half resolution.
*/
static void GetChromaSamples( unsigned char src[],
							  unsigned int cbPhase, unsigned int cbHF, unsigned int cbVF,
							  int saturation, int vsaturation )
{
	int x, i, c, s, u, v, lu, lv, fc, fs;

	for (x=0; x < decWidth/2; x++)
	{
		u = 0;
		v = 0;

		for (i=0; i < 2; i++)
		{
			// cosine filter samples, extract color frequency, downsample
			c  =   704 * (src[ 0] - src[1]);
			c +=  1559 * (src[-1] - src[2]);
			c +=  1078 * (src[-2] - src[3]);
			c +=  -457 * (src[-3] - src[4]);
			c += -1020 * (src[-4] - src[5]);
			c +=  -282 * (src[-5] - src[6]);
			c +=   199 * (src[-6] - src[7]);
			c +=   141 * (src[-7] - src[8]);
			c >>= 4;

			// sine filter samples, extract color frequency, downsample
			s  =  1700 * (src[ 0] + src[1]);
			s +=   364 * (src[-1] + src[2]);
			s += -1525 * (src[-2] + src[3]);
			s += -2182 * (src[-3] + src[4]);
			s +=   821 * (src[-4] + src[5]);
			s +=   962 * (src[-5] + src[6]);
			s +=   199 * (src[-6] + src[7]);
			s +=  -340 * (src[-7] + src[8]);
			s >>= 4;

			// combine sine and cosine data at color frequency phase
			fs = cosint[(cbPhase - 0x40000000) >> 22];
			fc = cosint[ cbPhase			   >> 22];
			u += (fs*c + fc*s) >> 9;
			v += (fc*c - fs*s) >> 9;

			v = -v;
			src += rawWidth;
			cbPhase += cbVF;
		}
		src		+= 10 - 2*rawWidth;
		cbPhase	+= cbHF - 2*cbVF;

		u *= saturation;
		v *= vsaturation;
		lu = PvU[x];
		lv = PvV[x];
		PvU[x] = u;
		PvV[x] = v;

		// combine 2 lines to correct phase variance
		u = (u + lu) >> 22;
		v = (v + lv) >> 22;
		scanCbU[x] = (u < -128) ? -128 : ((u > 127) ? 127 : u);
		scanCrV[x] = (v < -128) ? -128 : ((v > 127) ? 127 : v);
	}
}


/*
 Get color burst information
*/
static void GetColorBurstData(register unsigned char *src, unsigned int *cbPhase,
  unsigned int *cbHF, unsigned int *cbVF, unsigned int *cbOdd )
{
	static int cbPhases[decHeight];
	double f_amp;
	unsigned int c0, c1, s0, s1, p0, p1, p2, p3, f_cb;
	int i, v, x, y, f_cos, f_sin;

 	// determine phase offsets of colorburst
	for (y=0; y < decHeight; y++)
	{
		for (x=0, f_cos=0, f_sin=0, f_cb=0; x < 72; x++)
		{
			i      = *src++;
			f_cos += cosint[ f_cb               >> 22] * i;
			f_sin += cosint[(f_cb - 0x40000000) >> 22] * i;
			f_cb  += FcHF0;
		}
		src  += (rawWidth - 72);
		double div = sqrt( (double) f_cos*f_cos + (double) f_sin*f_sin );
		if (div != 0)
			f_amp = 1 / div;
		else
      		f_amp = 10000;

		c1 = (unsigned int) (acos( f_cos * f_amp ) * divPi);
		s0 = (unsigned int) (asin( f_sin * f_amp ) * divPi) + 0x80000000;
		c0 = ~c1;
		s1 = ~s0 + 0x80000000;
		p0 = abs(c0 - s0);  p1 = abs(c0 - s1);
		p2 = abs(c1 - s0);  p3 = abs(c1 - s1);

		if (p0 < p2 && p0 < p3 || p1 < p2 && p1 < p3)  c1 = c0;
		if (p0 < p1 && p0 < p3 || p2 < p1 && p2 < p3)  s1 = s0;

		cbPhases[y] = (s1 >> 1) + (c1 >> 1);
	}

	// Vc polarity detection  (even/odd frame)
	for (i=2, p0=0, p1=0; i < (decHeight - 2); i += 2)
	{
		p0 += abs( cbPhases[i] - cbPhases[i - 1] ) >> 8;
		p1 += abs( cbPhases[i + 1] - cbPhases[i] ) >> 8;
	}
	*cbOdd = (p0 < p1);

	// sample rate correction (color burst)
	for (i=8, v=0; i < (decHeight /2); i++)
	{
		v += (cbPhases[(decHeight /2) + i] - cbPhases[i]) >> 8;
	}
	v			= ((v /(decHeight /2)) << 8) /((decHeight /2) - 8);
	*cbHF		= v /rawWidth +  FcHF0;
	*cbVF		= v           +  FcVF0;
	i			= cbPhases[0] +  174 * (*cbHF) - 0x60000000L;
	i		   -= (*cbOdd ? 0 : -0x40000000);
	*cbPhase	= i - *cbHF - (*cbHF >> 1);
}
